//
//  SMBuffer.c
//  Taken from FSStream
//
//  Created by Julien-Pierre Avrous on 22/08/06.
//  Copyright 2006 SourceMac.com. All rights reserved.
//

#include "SMBuffer.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* Init */
SMBuffer::SMBuffer()
{
	first=NULL;
	last=NULL;
	tsize=0;
	tcount=0;
		
	/* for copy full read */
	lastData=NULL;
	lastSize=0;
	lastPosition=0;
		
	pthread_mutex_init(&mutex, NULL);
}


/*
	Add data block to the foot of the buffer. Thread safe.
	You can release data after this call : he is copied.
*/
void SMBuffer::addBuffer(const void *data, unsigned int len)
{
	pthread_mutex_lock(&mutex);
	
	if(len>0 && data)
		{
		void *dt=malloc(len);
		if(dt)
			{
			memcpy(dt, data, len);
			_addBuffer(dt, len);
			}
		}
	
	pthread_mutex_unlock(&mutex);
}


/*
	Add data block to the foot of the buffer. Thread safe.
	If you use copyContinuousRead/pickUpContinuousRead, the data should be malloced
*/
void SMBuffer::addBufferNoCopy(const void *data, unsigned int len)
{
	pthread_mutex_lock(&mutex);
	
	if(len>0 && data)
		_addBuffer(data, len);
	
	pthread_mutex_unlock(&mutex);
}


/*
	Get data block at the buffer head. Thread Safe.
	data and length are filled by the block data and block length.
	You should clean data with a call to free() function
*/
 void SMBuffer::getBuffer(void **data, unsigned int *len)
{
	pthread_mutex_lock(&mutex);
	
	_getBuffer(data, len);
	
	pthread_mutex_unlock(&mutex);
}


/*
	This function see the buffer just like a continuous amount of data and not like a data
	block array like getBuffer. Thread Safe.
	outBuffer is the buffer to fill and wantedSize is data length that you want to get.
	The copied data is automaticaly released and deleted after the call to this function.
*/
unsigned int SMBuffer::copyContinuousRead(void *outBuffer, unsigned int len)
{
	void *aData;
	unsigned int aSize;
	char *cdataIn, *cdataOut=(char*)outBuffer;
	int i=0, j=0, c1, min;
	
	if(len<=0)
		return 0;
	
	pthread_mutex_lock(&mutex);
	
	/* previous buffer not empty */
	if(lastData)
		{
		cdataIn=(char*)lastData;
		c1=lastSize-lastPosition;
		min=c1<len?c1:len;
		
		memcpy(cdataOut, cdataIn+lastPosition, min);
		tsize-=min;
		lastPosition+=min;
		i+=min;
		
		if(lastPosition>=lastSize) //on a finit l'ancien buffer : on le libere
			{
			free(lastData);
			lastData=NULL;
			}
		if(min>=len)
			{ pthread_mutex_unlock(&mutex); return len; }
		}
	
	/* continue buffer */
	aData=NULL;
	aSize=0;
	min=0;
	_getBuffer(&aData, &aSize);
	tsize+=aSize;
	cdataIn=(char*)aData;
	j=0;
	while(i<len && aData)
		{
		min=(aSize<len-i)?aSize:len-i;
		memcpy(cdataOut+i, cdataIn, min);
		tsize-=min;
		i+=min;
		
		if(i<len)
			{
			free(aData);
			aData=NULL;
			aSize=0;
			_getBuffer(&aData, &aSize);
			tsize+=aSize;
			cdataIn=(char*)aData;
			}
		}
	if(aSize==min)
		{
		free(aData);
		aData=NULL;
		aSize=0;
		lastPosition=0;
		}
	else
		lastPosition=min;
	lastData=aData;
	lastSize=aSize;

	pthread_mutex_unlock(&mutex);
	
	return i;
}


/*
	This function see the buffer just like a continuous amount of data and not like a data
	block array like getBuffer. Thread safe.
	outBuffer is the buffer to fill and wantedSize is data length that you want to get.
	The copied data are not released and not deleted from the buffer. 
*/
unsigned int SMBuffer::pickUpContinuousRead(void *outBuffer, unsigned int len)
{
	void *aData;
	int aSize;
	char *cdataIn, *cdataOut=(char*)outBuffer;
	int i=0, j=0, c1, min;
	keyedBuffer *buf;
	
	if(len<=0)
		return 0;
	
	pthread_mutex_lock(&mutex);
	
	/* previous buffer not empty */
	if(lastData)
		{
		cdataIn=(char*)lastData;
		c1=lastSize-lastPosition;
		min=c1<len?c1:len;
		
		memcpy(cdataOut, cdataIn+lastPosition, min);
		i+=min;

		if(min>=len)
			{ pthread_mutex_unlock(&mutex); return len; }
		}
	
	/* continue buffer */
	aData=NULL;
	aSize=0;
	min=0;
	buf=last;
	if(buf)
		{
		aData=buf->data;
		aSize=buf->dataLen;
		buf=buf->prev;
		}
	cdataIn=(char*)aData;
	j=0;
	while(i<len && aData)
		{
		min=(aSize<len-i)?aSize:len-i;
		memcpy(cdataOut+i, cdataIn, min);
		i+=min;
		
		if(i<len)
			{
			aData=NULL;
			aSize=0;
			if(buf)
				{
				aData=buf->data;
				aSize=buf->dataLen;
				buf=buf->prev;
				}
			cdataIn=(char*)aData;
			}
		}

	pthread_mutex_unlock(&mutex);
	
	return i;
}


/*
	Just know if the buffer is empty or not
*/
bool SMBuffer::thereIsData()
{
	return first && last;
}


/*
	Just know the amount of data in the buffer
*/
unsigned int SMBuffer::bufferSize()
{
	return tsize;
}


/*
	Just know the number of frame in the buffer (one frame = one addBuffer)
*/
unsigned int SMBuffer::bufferFrameCount()
{
	return tcount;
}


/*
	Just clean the data stored in the buffer
*/
void SMBuffer::clean()
{
	pthread_mutex_lock(&mutex);
	
	void *data=NULL;
	keyedBuffer *buf=last, *t;
	
	while(buf)
		{
		data=buf->data;
		if(data) free(data);
		t=buf;
		buf=buf->prev;
		free(t);
		}
	last=NULL;
	first=NULL;
	
	if(lastData)
		free(lastData);
	lastData=NULL;
	lastPosition=0;
	lastSize=0;
	tsize=0;
	tcount=0;
	
	pthread_mutex_unlock(&mutex);
}





/* HERE PRIVATE FUNCTIONS : DO NOT USE OUT, THERE ARE NOT THREAD SAFE */



/*
	The add buffer private routine : here, no thread safe
*/
void SMBuffer::_addBuffer(const void *data, unsigned int len)
{
	keyedBuffer *dt=(keyedBuffer*)malloc(sizeof(keyedBuffer));
	if(!dt)
		return;
	
	dt->data=(void*)data;
	dt->dataLen=len;
	tsize+=len;
	tcount++;
	
	dt->next=first;
	dt->prev=NULL;
	
	if(!first && !last)
		last=dt;
	else
		first->prev=dt;
		
	first=dt;
}


/*
	The get buffer private routine : here, no thread safe
*/
void SMBuffer::_getBuffer(void **data, unsigned int *len)
{
	keyedBuffer *p;
	
	if(lastData && (!first || !last))
		{
		void *tdata=malloc(lastSize-lastPosition);
		memcpy(tdata, (char*)lastData+lastPosition, lastSize-lastPosition);
		free(lastData); lastData=NULL;
		
		*data=tdata;
		if(len) *len=lastSize-lastPosition;
		tsize-=(lastSize-lastPosition);
		}
	
	if(!first || !last)
		return;
	
	if(lastData)
		{
		void *tdata=malloc(last->dataLen+lastSize-lastPosition);
		
		memcpy(tdata, (char*)lastData+lastPosition, lastSize-lastPosition);
		free(lastData); lastData=NULL;
		
		memcpy((char*)tdata+lastSize-lastPosition, last->data, last->dataLen);
		free(last->data);
		
		*data=tdata;
		if(len) *len=lastSize+last->dataLen-lastPosition;
		tsize-=(lastSize+last->dataLen-lastPosition);
		}
	else
		{
		*data=last->data;
		if(len) *len=last->dataLen;
		tsize-=last->dataLen;
		}
	p=last->prev;
	
	if(first!=last)
		last->prev->next=NULL;
	else
		first=NULL;
	
	free(last);
	last=p;
	tcount--;
}




/*
	Instance clean
*/
SMBuffer::~SMBuffer()
{
	clean();
	
	pthread_mutex_destroy(&mutex);
}